import { useMemo, useCallback } from "react";
import { Store, useStore } from "@tanstack/react-store";
import { once, isEqual } from "lodash-es";
import { getRenderingType, isFancyTransition, isClient, useSmartValue, fancyParamsTransitioning } from "./utils.js";
import { encodeParam, decodeParam } from "./encoding.js";
import { useContextApi } from "./use-api.js";
import { withBatch, defaultApi } from "./api.js";
const createFancyParams = () => {
  let paramsStore = null;
  let api = null;
  const getLatestParams = (api2) => {
    const params = new URLSearchParams(api2.getSearch());
    const paramsMap = {};
    for (const [key, value] of params) {
      paramsMap[key] = value;
    }
    return paramsMap;
  };
  const init = once((contextApi, isClient2) => {
    api = withBatch(contextApi ?? defaultApi(getRenderingType()));
    api.registerListener((state) => {
      if (isFancyTransition(state)) {
        return;
      }
      paramsStore.setState(() => getLatestParams(api));
    });
    if (isClient2) {
      paramsStore = new Store({});
      paramsStore.state = getLatestParams(api);
    }
  });
  const decodeWithDefault = (value, defaultValue, options) => {
    var _a, _b;
    try {
      if (value === void 0) {
        return defaultValue;
      }
      const decoded = decodeParam(value, options.decode);
      const isValid = options.validate ? options.validate(decoded) : true;
      if (!isValid) {
        (_a = options.onError) == null ? void 0 : _a.call(options, value);
        return defaultValue;
      }
      return decoded;
    } catch (_e) {
      (_b = options.onError) == null ? void 0 : _b.call(options, value);
      return defaultValue;
    }
  };
  const useFancyParamGet = (paramName, options) => {
    const contextApi = useContextApi();
    init(contextApi, isClient);
    const { defaultValue } = options;
    const smartDefaultValue = useSmartValue(defaultValue);
    const value = isClient ? useStore(paramsStore, (s) => s[paramName]) : getLatestParams(api)[paramName];
    const decodedValue = useMemo(() => decodeWithDefault(value, smartDefaultValue, options), [value, smartDefaultValue, options]);
    return useSmartValue(decodedValue);
  };
  const useFancyParamSet = (paramName, options) => {
    if (!isClient) {
      throw new Error("Cannot use fancy params on the server");
    }
    const contextApi = useContextApi();
    init(contextApi, isClient);
    const { defaultValue } = options;
    const smartDefaultValue = useSmartValue(defaultValue);
    return useCallback((value, state) => {
      const { updateType = "replaceIn", encode } = options;
      const currentValue = paramsStore.state[paramName];
      const internalValue = typeof value === "function" ? value(decodeWithDefault(currentValue, smartDefaultValue, options)) : value;
      const newValue = encodeParam(internalValue, encode);
      if (currentValue === newValue || currentValue === void 0 && isEqual(smartDefaultValue, value)) {
        return;
      }
      paramsStore.setState((s) => {
        return { ...s, [paramName]: newValue };
      });
      const search = api.getSearch();
      const clearAll = !updateType.endsWith("In");
      const isReplace = updateType.startsWith("replace");
      const searchParams = !clearAll ? new URLSearchParams(search) : new URLSearchParams();
      if (!isEqual(smartDefaultValue, value)) {
        searchParams.set(paramName, newValue);
      } else {
        searchParams.delete(paramName);
      }
      const newHref = `?${searchParams.toString()}`;
      const internalState = { [fancyParamsTransitioning]: true, ...state };
      if (isReplace) {
        api.replaceState(newHref, internalState);
      } else {
        api.pushState(newHref, internalState);
      }
    }, [paramName, smartDefaultValue, options]);
  };
  const useFancyParam = (paramName, options) => {
    const value = useFancyParamGet(paramName, options);
    const setter = useFancyParamSet(paramName, options);
    return [value, setter];
  };
  return {
    useFancyParamGet,
    useFancyParamSet,
    useFancyParam,
    getInternals: () => {
      return {
        paramsStore,
        api
      };
    }
  };
};
export {
  createFancyParams
};
//# sourceMappingURL=core.js.map
