import { createFancyParams } from "./core.js";
import { camelCaseToKebab, useMemoOptions } from "./utils.js";
import { createWrappers } from "./wrappers.js";
import { encodeParam, createListCoder, defaultEncoder, defaultDecoder, stringCoder, numberCoder, booleanCoder, datetimeCoder, dateCoder } from "./encoding.js";
const typeBasedOptions = {
  string: {
    ...stringCoder
  },
  number: {
    ...numberCoder
  },
  boolean: {
    ...booleanCoder
  },
  datetime: {
    ...datetimeCoder
  },
  date: {
    ...dateCoder
  }
};
class InternalBuilder {
  constructor(state) {
    this.state = state;
  }
  withCoder(coder) {
    this.state = {
      ...this.state,
      ...coder
    };
    return this;
  }
  validate(validator) {
    this.state = {
      ...this.state,
      ...validator
    };
    return this;
  }
  withDefault(value) {
    this.state = {
      ...this.state,
      defaultValue: value
    };
    return this;
  }
  asDialog() {
    return {
      state: {
        ...this.state,
        type: "dialog"
      }
    };
  }
}
const f = new Proxy({}, {
  get(_target, prop) {
    return (value) => {
      const key = prop;
      const options = typeBasedOptions[key] || {};
      const listOptions = key === "list" ? createListCoder(value) : {};
      const defaultCoders = {
        encode: defaultEncoder,
        decode: defaultDecoder
      };
      const state = {
        onError: () => {
        },
        validate: () => true,
        type: "value",
        defaultValue: null,
        updateType: "replaceIn",
        prefix: "",
        ...defaultCoders,
        ...options,
        ...value || {},
        ...listOptions
      };
      return new InternalBuilder(state);
    };
  }
});
function create(schema, globalOptions) {
  const params = createFancyParams();
  const wrappers = createWrappers(params);
  const proxyWrapper = new Proxy({}, {
    get(_target, prop) {
      if (prop === "build") {
        return (values) => {
          const params2 = Object.entries(values).map(([key, value]) => {
            const schemaKey = camelCaseToKebab(key);
            const builder2 = schema[schemaKey];
            const { defaultValue, encode } = builder2.state;
            if (value === defaultValue) {
              return void 0;
            }
            const encodedValue = encodeParam(value, encode);
            return `${key}=${encodedValue}`;
          }).filter(Boolean);
          return params2.join("&");
        };
      }
      if (prop === "batch") {
        return (fn) => {
          const { paramsStore, api } = params.getInternals();
          paramsStore.batch(() => {
            api.batch(fn);
          });
        };
      }
      const schemaField = camelCaseToKebab(prop);
      const builder = schema[schemaField];
      const options = builder.state;
      const simpleMap = () => {
        const typedOptions = options;
        return {
          useSet: (options2) => {
            const memoedOptions = useMemoOptions(typedOptions, globalOptions, options2);
            return params.useFancyParamSet(prop, memoedOptions);
          },
          use: (options2) => {
            const memoedOptions = useMemoOptions(typedOptions, globalOptions, options2);
            return params.useFancyParam(prop, memoedOptions);
          }
        };
      };
      if (options.type === "value") {
        return simpleMap();
      } else if (options.type === "dialog") {
        const typedOptionsWithDialogDefault = {
          ...options,
          defaultValue: {
            isOpen: false,
            state: options.defaultValue
          }
        };
        return {
          useSet: (options2) => {
            const memoedOptions = useMemoOptions(typedOptionsWithDialogDefault, globalOptions, options2);
            return wrappers.useDialogParamSet(prop, memoedOptions);
          },
          use: (options2) => {
            const memoedOptions = useMemoOptions(typedOptionsWithDialogDefault, globalOptions, options2);
            return wrappers.useDialogParam(prop, memoedOptions);
          }
        };
      }
      throw new Error("Invalid type");
    }
  });
  return proxyWrapper;
}
export {
  create,
  f
};
//# sourceMappingURL=fancy.js.map
