import {useCallback, useMemo} from "react";
import {Store, useStore} from "@tanstack/react-store";
import {isEqual, once} from "lodash-es";

import { ParamStore } from "./store";
import {fancyParamsTransitioning, getRenderingType, isClient, isFancyTransition, useSmartValue} from "./utils";
import { decodeParam, encodeParam } from "./encoding";

import type {API, OptionsWithDefault, Setter} from "./types";
import {useContextApi} from "./use-api";
import {BatchingApi, defaultApi, withBatch} from "./api";

export const createFancyParams = () => {
    let paramsStore: ParamStore = null!;
    let api: BatchingApi = null!;
    const getLatestParams = (api: API) => {
        const params = new URLSearchParams(api.getSearch());
        const paramsMap: Record<string, string> = {};
        for (const [key, value] of params) {
            paramsMap[key] = value;
        }
        return paramsMap;
    }

    const init = once((contextApi: API, isClient: boolean) => {
        api = withBatch(contextApi ?? defaultApi(getRenderingType()));
        api.registerListener((state: unknown) => {
            if (isFancyTransition(state)) {
                return;
            }
            paramsStore.setState(()=> getLatestParams(api))
        });
        if (isClient) {
            paramsStore = new Store<Record<string, string>>({});
            paramsStore.state = getLatestParams(api);
        }
    });

    const decodeWithDefault = <T,>(value: string | undefined, defaultValue: T, options: OptionsWithDefault<T>): T => {
        try {
            if (value === undefined) {
                return defaultValue
            }
            const decoded = decodeParam(value, options.decode)
            const isValid = options.validate ? options.validate(decoded) : true
            if (!isValid) {
                options.onError?.(value)
                return defaultValue
            }
            return decoded
        } catch (_e) {
            options.onError?.(value)
            return defaultValue
        }
    }

    const useFancyParamGet = <T,>(
        paramName: string,
        options: OptionsWithDefault<T>,
    ) => {
        const contextApi = useContextApi();
        init(contextApi, isClient);

        const {defaultValue} = options;
        const smartDefaultValue = useSmartValue(defaultValue);

        // eslint-disable-next-line react-hooks/rules-of-hooks
        const value = isClient ? useStore(paramsStore, (s) => s[paramName]) : getLatestParams(api)[paramName];
        const decodedValue = useMemo(() => decodeWithDefault(value, smartDefaultValue, options), [value, smartDefaultValue, options]);

        return useSmartValue(decodedValue) as T
    }

    const useFancyParamSet = <T,S extends object = object>(
        paramName: string,
        options: OptionsWithDefault<T>,
    ) => {
        if (!isClient) {
            throw new Error("Cannot use fancy params on the server");
        }
        const contextApi = useContextApi();
        init(contextApi, isClient);

        const {defaultValue} = options;
        const smartDefaultValue = useSmartValue(defaultValue);

        return useCallback((value: Setter<T>, state?: S) => {
            const {updateType = "replaceIn", encode} = options;
            const currentValue = paramsStore.state[paramName];
            const internalValue = typeof value === "function"
                ? ( value as (prev: T) => T )(decodeWithDefault(currentValue, smartDefaultValue, options))
                : value;

            const newValue = encodeParam(internalValue, encode);

            if (currentValue === newValue || (currentValue === undefined && isEqual(smartDefaultValue, value))) {
                return;
            }

            paramsStore.setState((s) => {
                return ({...s, [paramName]: newValue});
            });

            const search = api.getSearch();
            const clearAll = !updateType.endsWith("In");
            const isReplace = updateType.startsWith("replace");
            const searchParams = !clearAll ? new URLSearchParams(search) : new URLSearchParams();

            if (!isEqual(smartDefaultValue, value)) {
                searchParams.set(paramName, newValue);
            } else {
                searchParams.delete(paramName);
            }

            const newHref = `?${searchParams.toString()}`
            const internalState = {[fancyParamsTransitioning]: true, ...state};

            if (isReplace) {
                api.replaceState(newHref, internalState);
            } else {
                api.pushState(newHref, internalState);
            }
        }, [paramName, smartDefaultValue, options])
    }

    const useFancyParam = <T,>(
        paramName: string,
        options: OptionsWithDefault<T>,
    ) => {
        const value = useFancyParamGet<T>(paramName, options);
        const setter = useFancyParamSet<T>(paramName, options);
        return [value, setter] as const;
    }

    return {
        useFancyParamGet,
        useFancyParamSet,
        useFancyParam,
        getInternals: () => {
            return {
                paramsStore,
                api,
            }
        }
    }
}

export type FancyParamsCore = ReturnType<typeof createFancyParams>;