import {createFancyParams} from "./core";
import {
    BuildArg,
    Coder,
    DialogValue,
    FancyParams,
    FinalOptions,
    ListOptions,
    OptionsBuilder,
    RequiredOptions,
    Schema,
    UrlOptions,
    Validator
} from "./types";
import {camelCaseToKebab, useMemoOptions} from "./utils";
import {createWrappers} from "./wrappers";
import {
    booleanCoder,
    createListCoder,
    dateCoder,
    datetimeCoder,
    defaultDecoder,
    defaultEncoder,
    encodeParam,
    numberCoder,
    stringCoder
} from "./encoding";

type OptionsConfig = {
    [key in keyof OptionsBuilder]?: unknown
}

const typeBasedOptions: OptionsConfig =  {
    string: {
        ...stringCoder,
    },
    number: {
        ...numberCoder,
    },
    boolean: {
        ...booleanCoder,
    },
    datetime: {
        ...datetimeCoder,
    },
    date: {
        ...dateCoder,
    },
}

class InternalBuilder {
    state: RequiredOptions<unknown>
    constructor(state: RequiredOptions<unknown>) {
        this.state = state;
    }
    withCoder(coder: Coder<unknown>) {
        this.state = {
            ...this.state,
            ...coder,
        }
        return this;
    }
    validate(validator: Validator) {
        this.state = {
            ...this.state,
            ...validator,
        }
        return this;
    }
    withDefault(value: NonNullable<unknown>) {
        this.state = {
            ...this.state,
            defaultValue: value,
        }
        return this;
    }
    asDialog() {
        return {
            state: {
                ...this.state,
                type: "dialog",
            }
        }
    }
}

export const f = new Proxy({} as OptionsBuilder, {
    get(_target, prop: string) {
        return (value?: UrlOptions) => {
            const key = prop as keyof OptionsBuilder ;
            const options = typeBasedOptions[key] || {};

            const listOptions= key=== "list" ? createListCoder(value as ListOptions<unknown>) : {}
            const defaultCoders = {
                encode: defaultEncoder,
                decode: defaultDecoder,
            }

            const state = {
                onError: () => {},
                validate: () => true,
                type: "value" as const,
                defaultValue: null,
                updateType: "replaceIn",
                prefix: "",
                ...defaultCoders,
                ...options,
                ...( value || {}),
                ...listOptions
            } satisfies RequiredOptions<unknown>

            return new InternalBuilder(state);
        };
    },
});


export function create<T extends Schema>(schema: T, globalOptions?: UrlOptions): FancyParams<T> {
    const params = createFancyParams();
    const wrappers = createWrappers(params);

    const proxyWrapper = new Proxy({} as FancyParams<T>, {
        get(_target, prop: string) {
            if (prop === 'build') {
                return (values: BuildArg<T>) => {
                    const params = Object.entries(values).map(([key, value]) => {
                        const schemaKey = camelCaseToKebab(key);
                        const builder = schema[schemaKey] as unknown as InternalBuilder
                        const {defaultValue, encode} = builder.state;
                        if (value === defaultValue) {
                            return undefined
                        }
                        const encodedValue = encodeParam(value, encode);
                        return `${key}=${encodedValue}`;
                    }).filter(Boolean);
                    return params.join("&");
                }
            }
            if (prop === 'batch') {
                return (fn: () => void) => {
                    const {paramsStore, api} = params.getInternals();
                    paramsStore.batch(() => {
                        api.batch(fn);
                    });
                }
            }
            const schemaField = camelCaseToKebab(prop);
            const builder = schema[schemaField] as unknown as InternalBuilder;
            const options = builder.state;

            const simpleMap = <T,>() => {
                const typedOptions = options as RequiredOptions<T>;
                return {
                    useSet: (options?: FinalOptions) => {
                        const memoedOptions = useMemoOptions(typedOptions, globalOptions, options);
                        return params.useFancyParamSet(prop, memoedOptions);
                    },
                    use: (options?: FinalOptions) => {
                        const memoedOptions = useMemoOptions(typedOptions, globalOptions, options);
                        return params.useFancyParam(prop, memoedOptions);
                    },
                }
            }

            if (options.type === "value") {
                return simpleMap<unknown>();
            } else if (options.type === "dialog") {
                const typedOptionsWithDialogDefault = {
                    ...options,
                    defaultValue: {
                        isOpen: false,
                        state: options.defaultValue,
                    }
                } as RequiredOptions<DialogValue<unknown>>;
                return {
                    useSet: (options?: FinalOptions) => {
                        const memoedOptions = useMemoOptions(typedOptionsWithDialogDefault, globalOptions, options);
                        return wrappers.useDialogParamSet(prop, memoedOptions);
                    },
                    use: (options?: FinalOptions) => {
                        const memoedOptions = useMemoOptions(typedOptionsWithDialogDefault, globalOptions, options);
                        return wrappers.useDialogParam(prop, memoedOptions);
                    }
                }
            }
            throw new Error("Invalid type");
        }
    });
    return proxyWrapper;
}